#include <stdio.h>
#include <gccore.h>
#include <ogcsys.h>
#include <ogc/dvd.h>
#include <ogc/es.h>
#include <ogc/ipc.h>
#include <ogc/ios.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <malloc.h>
#include <ogcsys.h>
#include <stdarg.h>

#include "gcbackuplauncher_dol.h"
#include "background_rnc.h"

#define GC_HOMEBREW_ADDR 0x80800000
#define BACKGROUND_SIZE 563200

#define CIOS_VERSION 249

#define BC 		0x0000000100000100ULL
#define MIOS 	0x0000000100000101ULL

#define IOCTL_DI_READID				0x70
#define IOCTL_DI_READ				0x71
#define IOCTL_DI_RESET				0x8A
#define IOCTL_DI_READ_UNENCRYPTED	0xD0

#define GC_MAGIC_HOMEBREW "gchomebrew dol"

static GXRModeObj *vmode;
static u8 *gfxbuffer;
static u32 *xfb = NULL;

static int __dvd_fd = -1;
static int previously_initd = 0;
static char __di_fs[] ATTRIBUTE_ALIGN(32) = "/dev/di";

u8 dicommand[32] ATTRIBUTE_ALIGN(32);
u8 dibufferio[32] ATTRIBUTE_ALIGN(32);
static tikview view ATTRIBUTE_ALIGN(32);

void debug_printf(const char *fmt, ...)
{
	char buf[128];
	va_list ap;

#ifdef GEKKO_DEBUG
	int len;
	static char usbgecko_connected;
	static char usbgecko_checked;
#endif

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

#ifdef GEKKO_DEBUG
	if (!usbgecko_checked) {
		usbgecko_connected = usb_isgeckoalive(GEKKO_CHANNEL);
		usbgecko_checked = 1;
		usb_flush(GEKKO_CHANNEL);
	}
	if (usbgecko_connected) {
		len = strnlen(buf, sizeof(buf));
		usb_sendbuffer_safe(GEKKO_CHANNEL, buf, len);
	}
#endif
}

/* Synchronous DVD stuff.. bad! */

/* Open /dev/di */
int WiiDVD_Init()
{
	if (!previously_initd) {
		int ret;

		ret = IOS_Open(__di_fs, 0);
		if (ret < 0)
			return ret;
		__dvd_fd = ret;
		previously_initd = 1;
	}
	return 0;
}

/* Resets the drive, spins up the media */
void WiiDVD_Reset()
{
	memset(dicommand, 0, 32);
	dicommand[0] = IOCTL_DI_RESET;
	((u32 *) dicommand)[1] = 1;	//spinup(?)
	IOS_Ioctl(__dvd_fd, dicommand[0], &dicommand, 0x20, NULL, 0);
}

/* Read the Disc ID */
int WiiDVD_ReadID(void *dst)
{
	int ret;

	memset(dicommand, 0, 32);
	dicommand[0] = IOCTL_DI_READID;
	((u32 *) dicommand)[1] = 0;
	((u32 *) dicommand)[2] = 0;
	ret = IOS_Ioctl(__dvd_fd, dicommand[0], &dicommand, 0x20, dst, 0x20);
	return ret;
}

int WiiDVD_Read(void *dst, u32 size, u32 offset)
{
	int ret;

	memset(dicommand, 0, 32);
	dicommand[0] = IOCTL_DI_READ_UNENCRYPTED;
	((u32 *) dicommand)[1] = size;
	((u32 *) dicommand)[2] = offset >> 2;
	ret = IOS_Ioctl(__dvd_fd, dicommand[0], &dicommand, 0x20, dst, size);
	return ret;
}

static void Initialise(void)
{
	/* Load cIOS. */
	IOS_ReloadIOS(CIOS_VERSION);

	VIDEO_Init();

	vmode = VIDEO_GetPreferredMode(NULL);

	VIDEO_Configure(vmode);

	xfb = (u32 *) MEM_K0_TO_K1(SYS_AllocateFramebuffer(vmode));

	console_init(xfb, 50, 180, vmode->fbWidth, 480, vmode->fbWidth * 2);

	VIDEO_ClearFrameBuffer(vmode, xfb, COLOR_BLACK);
	VIDEO_SetNextFramebuffer(xfb);
	VIDEO_SetBlack(0);
	VIDEO_Flush();

	VIDEO_WaitVSync();
	if (vmode->viTVMode & VI_NON_INTERLACE)
		VIDEO_WaitVSync();

	/* Setup Graphics */
	gfxbuffer = (u8 *) memalign(32, BACKGROUND_SIZE);
	if (gfxbuffer == NULL) {
		printf("Failed to allocated graphics memory\n");
		return;
	}

	depackrnc1((u8 *) background_rnc, gfxbuffer);
	VIDEO_ClearFrameBuffer(vmode, xfb,
		*((unsigned long *) &gfxbuffer[BACKGROUND_SIZE - 4]));
	memcpy(xfb, gfxbuffer, BACKGROUND_SIZE);
}

void copy_program(void)
{
	u8 *data = (u8 *) GC_HOMEBREW_ADDR;

	memcpy(data, gcbackuplauncher_dol, gcbackuplauncher_dol_size);
	DCFlushRange(data, gcbackuplauncher_dol_size);
	ICInvalidateRange(data, gcbackuplauncher_dol_size);
}

void set_video_mode(int mode)
{
	/* Select MIOS video mode by disc mode. */
	switch (mode) {
	case 'E':
	case 'J':
		/* NTSC */
		vmode = &TVNtsc480IntDf;
		VIDEO_Configure(vmode);
		xfb = (u32 *) MEM_K0_TO_K1(SYS_AllocateFramebuffer(vmode));
		console_init(xfb, 50, 180, vmode->fbWidth, 480, vmode->fbWidth * 2);
		if (gfxbuffer != NULL) {
			VIDEO_ClearFrameBuffer(vmode, xfb,
				*((unsigned long *) &gfxbuffer[BACKGROUND_SIZE - 4]));
			memcpy(xfb, gfxbuffer, BACKGROUND_SIZE);
		} else {
			VIDEO_ClearFrameBuffer(vmode, xfb, COLOR_BLACK);
		}
		break;

	case 'D':
	case 'P':
	case 'F':
	case 'X':
	case 'Y':
	case 'U':
		/* PAL */
		vmode = &TVPal528IntDf;
		VIDEO_Configure(vmode);
		xfb = (u32 *) MEM_K0_TO_K1(SYS_AllocateFramebuffer(vmode));
		console_init(xfb, 50, 180, vmode->fbWidth, 528, vmode->fbWidth * 2);
		VIDEO_ClearFrameBuffer(vmode, xfb, COLOR_BLACK);
		if (gfxbuffer != NULL) {
			VIDEO_ClearFrameBuffer(vmode, xfb,
				*((unsigned long *) &gfxbuffer[BACKGROUND_SIZE - 4]));
			memcpy(xfb, gfxbuffer, BACKGROUND_SIZE);
		} else {
			VIDEO_ClearFrameBuffer(vmode, xfb, COLOR_BLACK);
		}
		break;
	default:
			/* Unknown mode. Don't change. */
			break;
	}
	free(gfxbuffer);
	gfxbuffer = NULL;

	VIDEO_SetNextFramebuffer(xfb);
	VIDEO_SetBlack(0);
	VIDEO_Flush();

	VIDEO_WaitVSync();
	if (vmode->viTVMode & VI_NON_INTERLACE)
		VIDEO_WaitVSync();
}

/****************************************************************************
* Main
****************************************************************************/

int main()
{
	char *magic = (char *) 0x807FFFE0;
	unsigned char *discid = (unsigned char *) 0x80000000;

	Initialise();

	debug_printf("\n\n\n\n\nStart GC Backup Launcher by WiiGator\n\n\n");

	debug_printf("Opening /dev/di/ ..\n");
	WiiDVD_Init();

	debug_printf("Resetting DI interface ..\n");
	WiiDVD_Reset();

	debug_printf("Reading Disc ID ..\n");
	WiiDVD_ReadID((void *) discid);
	debug_printf("ID: %c%c%c%c\n", discid[0], discid[1], discid[2], discid[3]);
	DCFlushRange(discid, 32);

	set_video_mode(discid[3]);

	/* Copy included DOL, because this is faster when testing. Don't need to copy something on the SD card. */
	copy_program();

	debug_printf("Starting GC Backup Launcher...\n");

	/* Set magic value, to inform gc side about homebrew. */
	strcpy(magic, GC_MAGIC_HOMEBREW);
	DCFlushRange(magic, 32);
	ICInvalidateRange(magic, 32);

	*(volatile unsigned int *) 0xCC003024 |= 7;

	int retval = ES_GetTicketViews(BC, &view, 1);

	if (retval != 0)
		debug_printf("ES_GetTicketViews fail %d\n", retval);

	VIDEO_WaitVSync();
	if (vmode->viTVMode & VI_NON_INTERLACE) {
		VIDEO_WaitVSync();
	}

	retval = ES_LaunchTitle(BC, &view);
	debug_printf("ES_LaunchTitle fail %d\n", retval);
	return 0;
}
